import logging
import os
import subprocess
import re

class prodigal:
    def __init__(self, seq, outdir, ncores):
        self.seq =seq
        self.outdir = outdir
        self.logfile = os.path.join(outdir, "prodigal.log")
        if ncores == 1:
            logging.warning("You are running Prodigal with a single core. This will be slow. We recommend using 8-16 cores.")
        self.faa = self.run(ncores)
        self.bed = self.make_bed()

    def run(self, cores=1):
        logging.debug("Launching prodigal now: %s" % self.seq)
        co = os.path.join(self.outdir, "genecoord.bgk")
        faa = os.path.join(self.outdir, "prot.faa")
        lst = ["prodigal",
            "-i", self.seq,
            "-p", "meta",
           "-o", co,"-a", faa]
        try:
            # do not rerun for now if we already attempted the training once
            if not os.path.exists(faa):
                with open(self.logfile, "w") as fout:
                    subprocess.run(" ".join(lst), cwd=self.outdir, check=True, shell=True,
                                stdout = fout, stderr = fout)
            else:
                logging.debug("Prodigal output already exists")
        except Exception as e:
            logging.warning("Prodigal failed on this bin")
        return(faa)
    
    def make_bed(self):
        # parser for rpodigals faa using header information
        bedpath = os.path.join(self.outdir, "prot.bed")
        reg = re.compile("([\w\d.\-\+]+)_[0-9]+")
        with open(self.faa) as fin, open(bedpath, "w") as fout:
            for line in fin:
                if not line.startswith(">"):
                    continue
            
                segment = line.strip().split("#")
                # chromosome is >chrom_
                name = segment[0][1:].strip()

                m = reg.match(name)
                if m is not None:
                    chrom = m.group(1)
                else:
                    logging.warning("Could not extract chromsome name from protein header. This is a bug in pygmes. Please report this on our github.")
                    exit(1)
                
                # start 
                start = segment[1].strip()
                stop = segment[2].strip()
                strandinfo = segment[3].strip()
                if strandinfo == "1":
                    strand = "+"
                elif strandinfo == "-1":
                    strand = "-"
                else:
                    strand = "."
                
                newline = "{}\t{}\t{}\t{}\t{}\n".format(chrom, start, stop, strand, name)
                fout.write(newline)
                

        return bedpath

    def check_success(self):
        if os.stat(self.faa).st_size == 0:
            return False
        if not os.path.exists(self.faa):
            return False
        return True



import os 
import logging
import argparse
from pygmes.exec import gmes
from pygmes.diamond import multidiamond
import shutil
import gzip
from glob import glob
from pyfaidx import Fasta
import pygmes.version  as version
from pygmes.exec import create_dir
from pygmes.printlngs import write_lngs
from pygmes.prodigal import prodigal

this_dir, this_filename = os.path.split(__file__)
MODELS_PATH = os.path.join(this_dir, "data", "models")



class bin:
    def __init__(self, path, outdir):
        self.fasta = os.path.abspath(path)
        self.name = os.path.basename(path)
        self.outdir = os.path.join(os.path.abspath(outdir), self.name)
        create_dir(self.outdir)
        self.hybridfaa = None
    
    def get_best_faa(self):
        if self.kingdom is not None and self.kingdom in ["bacteria", "archaea"]:
            if self.prodigal.check_success():
                return(self.prodigal.faa, self.prodigal.bed, self.name, "prodigal")
        elif  hasattr(self, 'gmes') and self.gmes.check_success():
            # check if we made a hybrid
            if self.hybridfaa is not None and os.path.exists(self.hybridfaa):
                return(self.hybridfaa, self.hybridbed, self.name, "hybrid")
            else:
                return(self.gmes.finalfaa, self.gmes.bedfile, self.name, "GeneMark-ES")
        elif hasattr(self, "gmes") and not self.gmes.check_success():
            # this bin is euakryotic maybe, but gmes has failed. 
            # so we will return the prodigal peptides instead, eeventhough we 
            # know this might be of low quality
            if self.prodigal.check_success():
                return(self.prodigal.faa, self.prodigal.bed, self.name, "prodigal")
        # default to none, as we dont have proteins it seems
        logging.debug("No final faa for bin: %s" % self.name)
        return(None, None, self.name, None)

    def gmes_training(self, ncores = 1):
        outdir = os.path.join(self.outdir, "gmes_training")
        self.gmes = gmes(self.fasta, outdir, ncores)
        self.gmes.selftraining()
    
    def run_prodigal(self, ncores = 1, outdir=None):
        if outdir is None:
            outdir = os.path.join(self.outdir, "prodigal")
            create_dir(outdir)
        self.prodigal = prodigal(self.fasta,  outdir, ncores)

    def make_hybrid_faa(self, gmesfirst = True):
        logging.debug("Making a hybrid of bin %s" % self.name)
        try:
            if not self.gmes.check_success() or not self.prodigal.check_success():
                return
        except AttributeError:
            return
        outdir = os.path.join(self.outdir, "hybrid")
        create_dir(outdir)
        self.hybridfaa = os.path.join(outdir, "gmes_prodigal_merged.faa")
        self.hybridbed = os.path.join(outdir, "gmes_prodigal_merged.bed")

        def sane_faa(faa):
            if os.stat(faa).st_size == 0:
                return False
            try:
                fa = Fasta(faa)
                if len(fa.keys()) > 0:
                    return fa
                return False
            except Exception as e:
                print(e)
                logging.debug("Fasta has no entries: %s" % faa)
                return False

        def chromname(s):
            if s.startswith(">"):
                s = s[1:]
            return s.strip().split()[0].rsplit("_", 1)[0]

        def getchroms(fa):
            contigs = set()
            for seq in fa:
                contigs.add(chromname(seq.name))
            return contigs

        if gmesfirst:
            faa1 = self.gmes.finalfaa
            faa2 = self.prodigal.faa
            bed1 = self.gmes.bedfile
            bed2 = self.prodigal.bed
        else:
            faa2 = self.gmes.finalfaa
            faa1 = self.prodigal.faa
            bed2 = self.gmes.bedfile
            bed1 = self.prodigal.bed
        # load and check for valid fastas
        fa1  = sane_faa(faa1)
        fa2  = sane_faa(faa2)
        if fa1 is False or fa2 is False:
            return faa1

        # find contigs uniqly annotated in faa2
        contigs1 = getchroms(fa1)
        contigs2 = getchroms(fa2)
        leftover = contigs2 - contigs1
        if len(leftover) > 0:
            logging.debug("We found possible bacterial proteins in this proteome")
            # write prot from fa1
            with open(self.hybridfaa, "w") as fout:
                for seq in fa1:
                    fout.write(f">{seq.name}\n{seq}\n")
                # add new from fa2
                for seq in fa2:
                    # write to file
                    if chromname(seq.name) in leftover:
                        fout.write(f">{seq.name}\n{seq}\n")

            # make a merged bedfile
            shutil.copy(bed1, self.hybridbed)
            with open(self.hybridbed, "a") as fout, open(bed2) as fin:
                for line in fin:
                    if line.split("\t")[0] in leftover:
                        fout.write(line)

class pygmes:
    """
    Main class exposing the functionality

    Parameters:

    **fasta:** path to a fasta file

    **outdir:** path to a writable directory

    **db:** path to a diamond database with tax information

    **clean:** bool indicating if faster needs cleaning of headers

    **ncores:** number of threads to use
    """
    def __init__(self, fasta, outdir, db,  clean = True, ncores = 1):
        self.fasta = fasta
        self.outdir = outdir
        self.ncores = ncores

        if clean:
            # copy and clean file
            self.cleanfasta = self.clean_fasta(self.fasta, self.outdir)
        else:
            self.cleanfasta = self.fasta

        logging.info("Launching GeneMark-ES")
        g = gmes(self.cleanfasta, outdir, ncores)
        logging.debug("Run complete launch")
        g.run_complete(MODELS_PATH, db)
        if g.finalfaa:
            logging.debug("Copying final faa from: %s" % g.finalfaa)
            shutil.copy(g.finalfaa, os.path.join(self.outdir, "predicted_proteins.faa"))
            g.writetax()
        if g.bedfile:
            shutil.copy(g.bedfile, os.path.join(self.outdir, "predicted_proteins.bed"))
        else:
            logging.debug("Could not find bed file")
        print(g.bedfile)

    def clean_fasta(self, fastaIn, folder):
        create_dir(folder)
        name = os.path.basename(fastaIn)
        fastaOut = os.path.join(folder, name)
        if os.path.exists(fastaOut):
            logging.debug("Clean fasta %s exists" % name)
            return fastaOut
        mappingfile = os.path.join(folder, "mapping.csv")
        logging.debug("Cleaning fasta file")
        nms = []
        with open(fastaOut, "w") as o, open(mappingfile, "w") as mo:
            mo.write("old,new\n")
            if fastaIn.endswith(".gz"):
                openMethod = gzip.open
                gz = True
                logging.debug("reading gzipped file")
            else:
                openMethod = open
                gz = False
            # read in the fasta
            with openMethod(fastaIn) as f:
                for line in f:
                    if gz:
                        line = line.decode()
                    if line.startswith(">"):
                        line = line.strip()
                        l = line.split()
                        # get first element, usually a chromosome
                        N = l[0].strip()
                        # while the name is already taken, count up and reformat
                        n = N
                        i = 0
                        while n in nms:
                            n = "{}.{}".format(N, i)
                            i += 1
                        o.write("{}\n".format(n))
                        # finally add to know  name list
                        nms.append(n)
                        mappingline = "{},{}\n".format(line,n).replace(">", "")
                        mo.write(mappingline)
                    else:
                        o.write(line)
        return fastaOut
        
class metapygmes(pygmes):
    """
    run pygmes in metagenomic mode. This means
    we will first try the self training mode on
    all bins in the given folder
    We will then use the models from all runs
    to predict proteins in the remaining bins
    and choose the protein prediction with the largest
    number of AA. We then infer the lineage of each bin
    """
    def __init__(self, bindir, outdir, db, clean = True, ncores = 1, infertaxonomy = True, fill_bac_gaps = True):
        # find all files and 
        outdir = os.path.abspath(outdir)
        self.outdir = outdir
        bindir = os.path.abspath(bindir)
        fa = glob(os.path.join(bindir, "*.fa"))
        fna = glob(os.path.join(bindir, "*.fna"))
        fasta = glob(os.path.join(bindir, "*.fasta"))
        files = fa + fna + fasta
        prodigaldir = os.path.join(self.outdir, "prodigal")
        # convert all files to absolute paths
        files = [os.path.abspath(f) for f in files]
        names = [os.path.basename(f) for f  in files]
        if len(names) != len(set(names)):
            logging.warning("Bin files need to have unique names")
            exit(1)

        #outdirs = [os.path.join(outdir, name) for name in names]
        #proteinfiles = []
        #proteinnames = []
        # if needed, we clean the fasta files
    
        if clean:
            logging.info("Cleaning input fastas")
            cleanfastadir = os.path.join(outdir, "fasta_clean")
            files = [self.clean_fasta(f, cleanfastadir) for f in files ]


        # bin list to keep all the bins and handle all the operations
        binlst = []
        bindirs = os.path.join(outdir, "bins")
        for path in files:
            binlst.append(bin(path, bindirs))

        # run prodigal
        logging.info("Running prodigal on all bins")
        for b in binlst:
            b.run_prodigal(ncores = ncores)
        
        # now we can already get a first lineage estimation
        # diamond is faster when using more sequences
        # thus we pool all fasta together and seperate them afterwards
        diamonddir = os.path.join(outdir, "diamond", "step_1")
        create_dir(diamonddir)
        logging.info("Predicting the lineage")
        proteinfiles = [b.prodigal.faa for b in binlst if b.prodigal.check_success()]
        proteinnames = [b.name for b in binlst if b.prodigal.check_success()]
        dmnd_1 = multidiamond(proteinfiles, proteinnames, diamonddir, db = db, ncores = ncores)
        logging.debug("Ran diamond and inferred lineages")
        # assign a taxonomic kingdom based on the first lineage estimation
        anyeuks = False
        for b in binlst:
            b.first_lng_estimation = None
            b.kingdom = None
            if b.name in dmnd_1.lngs.keys():
                # as no lng was infered for this bin, we could try prodigal
                b.first_lng_estimation = dmnd_1.lngs[b.name]
                if 2 in b.first_lng_estimation['lng']:
                    b.kingdom = "bacteria"
                elif 2759 in b.first_lng_estimation['lng']:
                    b.kingdom = "eukaryote"
                    anyeuks = True
                elif 2157 in b.first_lng_estimation['lng']:
                    b.kingdom = "archaea"
                else:
                    anyeuks = True

        if anyeuks == False:
            logging.info("All bins are prokaryotes, we can skip the GeneMark-ES steps")
        else:
            # for all bins that are euakryotic or could not be assigned a lineage,
            # we try GeneMark-ES in a two step mode
            modeldir = os.path.join(outdir, "gmes_models")
            create_dir(modeldir)
            nmodels = 0
            logging.info("Running GeneMark-ES in self training")
            for b in binlst:
                if b.kingdom is None or b.kingdom == "eukaryote":
                    # run self training
                    b.gmes_training(ncores = ncores)
                    expectedmodel = os.path.join(b.gmes.outdir, "output","gmhmm.mod")
                    if os.path.exists(expectedmodel):
                        shutil.copy(expectedmodel, os.path.join(modeldir, "{}.mod".format(b.name)))
                        nmodels += 1
            # check if any bins were not predicted, if so we can use the models
            # from other bins to get a better estimate
            # if thats not possible, we could still run pygmes in non metagenomic 
            # on each bin, but that should be decied by the user
            if nmodels == 0:
                logging.debug("No models were successfully trained")
            else:
                for b in binlst:
                    if b.kingdom is None or b.kingdom == "eukaryote":
                        if b.gmes.check_success() is False:
                            b.gmes.premodel(modeldir)
                            # if successfull, overwrite the gmes, with the successfull gmes
                            if b.gmes.bestpremodel is not False and b.gmes.bestpremodel.check_success():
                                b.gmes = b.gmes.bestpremodel
            # now we have proteins predicted for all
            # we can now give each bin the chance to merge prodigal and Gmes predictions
            for b in binlst:
                b.make_hybrid_faa()

            # now we update the lineages using the new proteins
            # and then we can create a final set of protein files
            diamonddir = os.path.join(outdir, "diamond", "step_2")
            create_dir(diamonddir)
            proteinfiles = []
            proteinnames = []
            for b in binlst:
                path, bedpath, name, software = b.get_best_faa()
                if path is not None and b.kingdom not in ["bacteria", "archaea"]:
                    proteinfiles.append(path)
                    proteinnames.append(name)
            if len(proteinfiles) > 0:
                logging.info("Predicting the lineage using the results from GeneMark-ES")
                dmnd_2 = multidiamond(proteinfiles, proteinnames, diamonddir, db = db, ncores = ncores)
                for b in binlst:
                    if b.name in dmnd_2.lngs.keys():
                        # as no lng was infered for this bin, we could try prodigal
                        b.first_lng_estimation = dmnd_2.lngs[b.name]
                        if 2 in b.first_lng_estimation['lng']:
                            b.kingdom = "bacteria"
                        elif 2759 in b.first_lng_estimation['lng']:
                            b.kingdom = "eukaryote"
                        elif 2157 in b.first_lng_estimation['lng']:
                            b.kingdom = "archaea"
            else:
                logging.info("No changes after applying GeneMark-ES")

        # now we can make a final FAA folder:
        finaloutdir = os.path.join(self.outdir, "predicted_proteomes")
        finalbeddir = os.path.join(finaloutdir, "bed")
        create_dir(finaloutdir)
        create_dir(finalbeddir)
        lngs = {}
        metadataf = os.path.join(self.outdir, "metadata.tsv")
        metadata = {}
        finalfaas = {}
        for b in binlst:
            t =  os.path.join(finaloutdir, "{}.faa".format(b.name))
            bt = os.path.join(finalbeddir, "{}.bed".format(b.name))
            path, bedpath, name, software = b.get_best_faa()
            b.software = software
            metadata[b.name] = {"path": path, 
                             "software": software, 
                             "nprot": None,
                             "lng": [],
                             "name": b.name}
            if path is not None:
                shutil.copy(path, t)
                shutil.copy(bedpath, bt)
                metadata[b.name]['path'] = t
                finalfaas[b.name] = {"faa": t, "fasta": b.fasta}
                try:
                    fa = Fasta(t)
                    metadata[b.name]['nprot'] = len(fa.keys())
                except Exception as e:
                    print(e)
                    metadata[b.name]['nprot'] = 0
            if b.first_lng_estimation is not None:
                lngs[b.name] = {}
                lngs[b.name]['lng'] = b.first_lng_estimation['lng']
                lngs[b.name]['n'] = b.first_lng_estimation['n']
                metadata[b.name]['lng'] = "-".join([str(x) for x in b.first_lng_estimation['lng']])
        logging.debug("Copied files, now writing lineages")
        lngfile = os.path.join(outdir, "lineages.tsv")
        write_lngs(lngs, lngfile)
        # write metadata to disk
        logging.debug("Writing metadata")
        with open(metadataf, "w") as fout:
            keys = ['name', "path", "software", "nprot", "lng"] 
            fout.write("\t".join(keys))
            fout.write("\n")
            for k,v in metadata.items():
                l = []
                for key in keys:
                    l.append(str(v[key]))
                fout.write("\t".join(l))
                fout.write("\n")
        
        # make a single file for CAT, including a prefix so CAT will not get confused
        def single_fasta(fastas, names, output, sep = "_"):
            if len(fastas) != len(names):
                logging.warning("Number of Fastas does not match names")
                exit(1)
            nseqs = 0
            with open (output, "w") as fout:
                for fasta, name in zip(fastas, names):
                    with open(fasta) as fin:
                        for line in fin:
                            if line.startswith(">"):
                                line = line.strip().split()[0][1:]
                                line = ">{}{}{}\n".format(name, sep, line)
                                nseqs += 1
                            fout.write(line)
            if nseqs == 0:
                logging.warning("No sequence in aggregate")
                exit(1)
        
        # make massive protein file:
        catdir = os.path.join(outdir, "CAT")
        create_dir(catdir)
        catfaa = os.path.join(catdir, "cat.faa")
        catfna = os.path.join(catdir, "cat.fna")
        names = list(finalfaas.keys())
        names.sort()
        faas = [finalfaas[name]['faa'] for name in names]
        fnas = [finalfaas[name]['fasta'] for name in names]
        single_fasta(fnas, names, catfna)
        single_fasta(faas, names, catfaa)




        logging.info("Successfully ran pygmes --meta")


def main():
    parser = argparse.ArgumentParser(description="Evaluate completeness and contamination of a MAG.")
    parser.add_argument("--input", "-i", type=str, help="path to the fasta file, or in metagenome mode path to bin folder")
    parser.add_argument("--output", "-o", type=str, required=True, help="Path to the output folder")
    parser.add_argument("--db", "-d", type=str, required=True, help="Path to the diamond DB")
    parser.add_argument("--noclean", dest="noclean", default = True, action="store_false",required=False, help = "GeneMark-ES needs clean fasta headers and will fail if you dont proveide them. Set this flag if you don't want pygmes to clean your headers")
    parser.add_argument("--ncores", "-n", type=int, required=False, default = 1,
            help="Number of threads to use with GeneMark-ES and Diamond")
    parser.add_argument("--meta", dest="meta", action = "store_true", default=False, help = "Run in metaegnomic mode")
    parser.add_argument(
        "--quiet", "-q", dest="quiet", action="store_true", default=False, help="Silcence most output",
    )
    parser.add_argument(
        "--debug", action="store_true", default=False, help="Debug and thus ignore safety",
    )
    parser.add_argument("-v", "--version", action="version", version=f"pygmes version {version.__version__}")
    options = parser.parse_args()

    # define logging
    logLevel = logging.INFO
    if options.quiet:
        logLevel = logging.WARNING
    elif options.debug:
        logLevel = logging.DEBUG
    logging.basicConfig(
        format="%(asctime)s %(message)s", datefmt="%m/%d/%Y %H:%M:%S: ", level=logLevel,
    )

    # check if input is readable
    if not os.path.exists(options.input):
        logging.warning("Input file does not exist: %s" % options.input)
        exit()
    logging.info("Starting pygmes")
    logging.debug("Using fasta: %s" % options.input)
    logging.debug("Using %d threads" % options.ncores)

    if not options.meta:
        pygmes(options.input, options.output, options.db, clean = options.noclean,
            ncores = options.ncores)
    else:
        metapygmes(options.input, options.output, options.db, clean = options.noclean,
            ncores = options.ncores)





import logging
from ete3 import NCBITaxa
ncbi = NCBITaxa()

def compare_taxa(tax1, tax2):
    score = 0
    joined = []
    for a, b in zip(tax1, tax2):
        if a == b:
            score += 1
            joined.append(a)
        else:
            break
    return(score, joined)

def print_lngs(tax1, tax2):
    """
    Takes two list of numbers and draws a simple line diagram showing the divergence
    """
    # try guessing the line width
    try:
        import shutil
        s = shutil.get_terminal_size()
        linewidth = s.columns
    except:
        linewidth = 80
    
    
    def makestr(lst):
        return([str(i) for i in lst])
    
    tax1 = makestr(tax1)
    tax2 = makestr(tax2)
    score, joined = compare_taxa(tax1, tax2)
    # shorten
    joined = makestr(joined)
    tax1 = tax1[score:]
    tax2 = tax2[score:]
    
    # lables
    label1 = "Model Lng."
    label2 = "Shared    "
    label3 = "Bin Lng.  "

    root = "{} {}".format(label2, " ".join(joined))
    tax1r = " ".join(tax1)
    tax2r = " ".join(tax2)
    # check if a lineage or root is to long
    def shorten(lst):
        nw = [lst[0]]
        for e in lst[1:-3]:
            if e != "...":
                nw.append(e)
        nw.append("...")
        nw.append(lst[-1])
        
        return(nw)
    
    # check if we can shorten the branches
    while len(root) + len(tax1r) > linewidth or len(root) + len(tax2r) > linewidth:
        canremove = 0
        # shorten first the branches then the root
        if len(root) + len(tax1r) > linewidth and len(tax1) >= 3:
            # keep track of if there is room to shorten
            if len(tax1) <= 3 and tax1[1] == "...":
                canremove = canremove - 1
            else:
                tax1 = shorten(tax1)
        else:
            canremove = canremove - 1
                
        if len(root) + len(tax2r) > linewidth and len(tax2) >= 3:
            # keep track of if there is room to shorten
            if len(tax2) <= 3 and tax2[1] == "...":
                canremove = canremove - 1
            else:
                tax2 = shorten(tax2)
        else:
            canremove = canremove - 1
        
        tax1r = " ".join(tax1)
        tax2r = " ".join(tax2)
        if canremove <= -2:
            break
            
    # check if we need to shorten the root
     # check if we can shorten the branches
    while len(root) + len(tax1r) > linewidth or len(root) + len(tax2r) > linewidth:
        canremove = 0
        # shorten first the branches then the root
        if len(root) + len(tax1r) > linewidth and len(joined) >= 3:
            # keep track of if there is room to shorten
            if len(joined) <= 3 and joined[1] == "...":
                canremove = canremove - 1
            else:
                joined = shorten(joined)
        else:
            canremove = canremove - 1

              
        root = "{} {}".format(label2, " ".join(joined))
        if len(root) + len(tax2r) > linewidth and len(joined) >= 3:
            # keep track of if there is room to shorten
            if len(joined) <= 3 and joined[1] == "...":
                canremove = canremove - 1
            else:
                joined = shorten(joined)
        else:
            canremove = canremove - 1

        root = "{} {}".format(label2, " ".join(joined))
        
        if canremove <= -1:
            break
     
                
    
    space = [" "] * (len(root) - len(label1))
    space = "".join(space)
    longspace = "".join([" "] * (len(root)))
    
    #print dendrogram
    print("")
    print("Infered lineage compared to the model lineage:")
    if len(tax1r) > 0:
        print(label1, end ="")
        print(space, end = "  ")
        print(tax1r)
        print(longspace, end = "/\n")
    else:
        print(label1)
        print("")
    print(root)
    if len(tax2r) > 0:
        print(longspace, end = "\\\n")
        print(label3, end ="")
        print(space, end = "  ")
        print(tax2r)
    else:
        print("")
        print(label3)
    print("\n")

def write_lngs(lngs, outfile):
    """
    write infered taxonomy in a machine and human readble format
    """
    logging.info("Translating lineage")
    with open(outfile, "w") as fout:
        fout.write("bin\ttaxid\tncbi_rank\tncbi_name\tbasedon\n")
        for binname, lngi in lngs.items():
            lng = lngi['lng']
            nprots = lngi['n']
            nms = ncbi.get_taxid_translator(lng)
            ranks = ncbi.get_rank(lng)
            for taxid in lng:
                if taxid in nms.keys():
                    name = nms[taxid]
                else:
                    name = "unnamed"
                fout.write(f"{binname}\t{taxid}\t{ranks[taxid]}\t{name}\t{nprots}\n") 
        logging.info("Wrote lineage to %s" % outfile)



import logging
import os
import subprocess
from pyfaidx import Fasta
from random import sample
from collections import defaultdict
from ete3 import NCBITaxa

ncbi = NCBITaxa()


def majorityvote(lngs, fraction=0.6):
    if fraction <= 0.5:
        logging.warning("fraction must be larger than 0.5")
    if len(lngs) == 0:
        return []
    ml = max([len(v) for v in lngs])
    i = 0
    n = len(lngs)

    lng = []
    while i < ml:
        choices = defaultdict(int)
        for l in lngs:
            if len(l) <= i:
                continue
            choices[l[i]] += 1
        # decide if this is majority material
        if len(choices) > 0:
            best = max(choices, key=lambda key: choices[key])
            if choices[best] / n >= fraction:
                lng.append(best)
            else:
                break
            i += 1
        else:
            break
    return lng


class diamond:
    def __init__(self, faa, outdir, db, ncores=1, sample=100):
        self.faa = faa
        self.outdir = outdir
        self.db = db
        self.ncores = ncores
        self.outfile = os.path.join(self.outdir, "diamond.results.tsv")
        self.log = os.path.join(self.outdir, "diamond.log")
        self.lineages = {}
        if ncores == 1:
            logging.warning("You are running Diamond with a single core. This will be slow. We recommend using 8-16 cores.")
        # sample n proteins
        logging.info("Subsampeling %d proteins" % sample)
        self.samplefile = os.path.join(self.outdir, "diamond.query.faa")
        self.sample(self.samplefile, sample)

        # runa search
        logging.info("Running diamond blastp")
        self.search(self.outfile, self.samplefile)
        logging.debug("Parsing diamond output")
        self.parse_results(self.outfile)
        # infer lineages
        logging.debug("Inferring the lineage")
        proteinlngs = self.lineage_infer_protein(self.result)
        self.lineage  = self.vote_bin(proteinlngs)
        logging.debug("Finished the diamond step")

    def search(self, outfile, query):
        if not os.path.exists(outfile) or os.stat(outfile).st_size == 0:
            logging.info("Running diamond now")
            lst = [
                "diamond",
                "blastp",
                "--db",
                self.db,
                "-q",
                query,
                "-p",
                str(self.ncores),
                "--evalue",
                str(1e-20),
                "--max-target-seqs",
                "3",
                "--outfmt",
                "6",
                "qseqid",
                "sseqid",
                "pident",
                "evalue",
                "bitscore",
                "staxids",
                "-o",
                outfile,
            ]
            with open(self.log , "w") as fout:
                subprocess.run(lst, stderr = fout, stdout = fout)
            logging.debug("Ran diamond")
        else:
            logging.info("Diamond output already exists ")
            logging.debug("AT: %s" %  outfile)

    def sample(self, output, n=200):
        logging.debug("Sampeling %d proteins from %s" % (n, self.faa))
        try:
            faa = Fasta(self.faa)
        except ZeroDivisionError:
            logging.warning(
                "Could not read the faa file as it probably \n contains no sequence information. \n Check file: %s "
                % self.faa
            )
            return 0

        keys = faa.keys()
        if len(keys) > n:
            keys = sample(keys, n)
        with open(output, "a") as fout:
            for k in keys:
                fout.write(f">{k}\n{str(faa[k])}\n")

    def parse_results(self, result):
        r = defaultdict(list)
        with open(result) as f:
            for line in f:
                l = line.strip().split("\t")
                r[l[0]].append(l[5])
        self.result = r

    def inferlineage(self, tax):
        if tax in self.lineages.keys():
            return self.lineages[tax]
        else:
            try:
                self.lineages[tax] = ncbi.get_lineage(tax)
                return self.lineages[tax]
            except ValueError:
                print(f"Not able to fetch lineage for taxid {tax}")
                return []

    def lineage_infer_protein(self, result):
        prot = {}
        for protein, taxids in result.items():
            lngs = []
            for taxid in taxids:
                l = self.inferlineage(taxid)
                if len(l) > 0:
                    lngs.append(l)

            prot[protein] = majorityvote(lngs)
        return prot

    def vote_bin(self, proteinlngs):
        lngs = [lng for prot, lng in proteinlngs.items()]
        return majorityvote(lngs)

    
class multidiamond(diamond):
    def __init__(self,proteinfiles, names, outdir, db, ncores = 1, nsample = 200):
        self.outdir = os.path.abspath(outdir)
        self.files = proteinfiles
        self.names = names
        self.samplefile = os.path.join(outdir, "samplefile.faa")
        self.outfile = os.path.join(outdir, "diamond.result")
        self.log = os.path.join(self.outdir, "diamond.log")
        self.db = db
        self.ncores = ncores
        self.lineages = {}
        if ncores == 1:
            logging.warning("You are running Diamond with a single core. This will be slow. We recommend using 8-16 cores.")
        # sample
        with open(self.samplefile, "w") as f:
            f.write("")
        for fasta, name in zip(self.files, self.names):
            self.sample(fasta, name, self.samplefile)
        
        # then run 
        self.search(self.outfile, self.samplefile)
        self.result = self.parse_results(self.outfile)
        self.lngs = self.vote_bins(self.result)

    def sample(self, fasta, name, output, n=200):
        logging.debug("Sampeling %d proteins from %s" % (n, fasta))
        try:
            faa = Fasta(fasta)
        except ZeroDivisionError:
            logging.warning(
                "Could not read the faa file as it probably \n contains no sequence information. \n Check file: %s "
                % fasta
            )
            return 0

        keys = faa.keys()
        if len(keys) > n:
            keys = sample(keys, n)
        with open(output, "a") as fout:
            for k in keys:
                fout.write(f">{name}_binseperator_{k}\n{str(faa[k])}\n")
    
    def parse_results(self, result):
        def subdict():
            return(defaultdict(list))
        r = defaultdict(subdict)
        with open(result) as f:
            for line in f:
                l = line.strip().split("\t")
                names = l[0].split("_binseperator_")
                binname = names[0]
                protein = names[1]
                r[binname][protein].append(l[5])
        return r
    
    def vote_bins(self, result):
        binnames = result.keys()
        lngs = {}
        for bin in binnames:
            protlng = self.lineage_infer_protein(result[bin])
            lngs[bin] = {}
            lngs[bin]["lng"] = self.vote_bin(protlng)
            lngs[bin]["n"] = len(protlng)

        return(lngs)



import logging
import os
import subprocess
import glob
import re
from pyfaidx import Fasta
from pyfaidx import FastaIndexingError
from random import sample
from collections import defaultdict
from pygmes.diamond import diamond
from pygmes.printlngs import print_lngs
from ete3 import NCBITaxa
import shutil

ncbi = NCBITaxa()


import urllib.request


url = "http://paulsaary.de/gmes/"
def create_dir(d):
    if not os.path.isdir(d):
        try:
            os.makedirs(d)
        except OSError as e:
            logging.warning(f"Could not create dir: {d}\n{e}")


def delete_folder(d):
    if os.path.exists(d):
        if os.path.isdir(d):
            try:
                shutil.rmtree(d)
            except Exception as e:
                logging.warning("Could not delete folder: %s" % d)
                print(e)

def touch(fname, mode=0o666, dir_fd=None, **kwargs):
    flags = os.O_CREAT | os.O_APPEND
    with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
        os.utime(f.fileno() if os.utime in os.supports_fd else fname,
            dir_fd=None if os.supports_fd else dir_fd, **kwargs)

class gmes:
    def __init__(self, fasta, outdir, ncores=1):
        self.fasta = os.path.abspath(fasta)
        self.outdir = os.path.abspath(outdir)
        self.logfile = os.path.join(self.outdir, "pygmes.log")
        self.loggtf = os.path.join(self.outdir, "pygmes_gtf.log")
        # make sure the output folder exists
        create_dir(self.outdir)
        self.ncores = ncores

        self.gtf = os.path.join(self.outdir, "genemark.gtf")
        self.protfaa = os.path.join(self.outdir, "prot_seq.faa")
        self.finalfaa = False
        self.finalgtf = False
        self.bedfile = False
        self.tax = []
        self.modelinfomap = {}
        if ncores == 1:
            logging.warning("You are running GeneMark-ES with a single core. This will be slow. We recommend using 8-16 cores.")

    def selftraining(self):
        failpath = os.path.join(self.outdir, "tried_already")
        if os.path.exists(failpath):
            logging.info("Self-training skipped, as we did this before and it failed")
            self.gtf2faa()
            return
        if os.path.exists(self.gtf):
            logging.info("GTF file already exists, skipping")
            self.gtf2faa()
            return

        logging.debug("Starting self-training")
        lst = [
            "gmes_petap.pl",
            "--v",
            "--fungus",
            "--ES",
            "--cores",
            str(self.ncores),
            "--min_contig",
            "5000",
            "--sequence",
            self.fasta,
        ]
        try:
            with open(self.logfile, "a") as fout:
                subprocess.run(" ".join(lst), cwd=self.outdir, check=True, shell=True,
                            stdout = fout, stderr = fout)
        except subprocess.CalledProcessError:
            touch(failpath)
            logging.info("GeneMark-ES in self-training mode has failed")
        # predict and then clean
        self.gtf2faa()
        self.clean_gmes_files()

    def clean_gmes_files(self):
        # clean if there are files to clean
        # this just keeps the foodprint lower
        rmfolders = ['run', 'info', 'data', 'output/data', 'output/gmhmm']
        for folder in rmfolders:
            p = os.path.join(self.outdir, folder)
            delete_folder(p)


    def prediction(self, model):
        self.model = model
        self.modelname = os.path.basename(model).replace(".mod","")
        failpath = os.path.join(self.outdir, "tried_already")
        if os.path.exists(failpath):
            logging.info("Prediction skipped, as we did this before and it failed")
            return
        if os.path.exists(self.gtf):
            logging.debug("GTF file already exists, skipping")
            self.gtf2faa()
            return
        logging.debug("Starting prediction")
        lst = [
            "gmes_petap.pl",
            "--v",
            "--predict_with",
            model,
            "--cores",
            str(self.ncores),
            "--sequence",
            self.fasta,
        ]
        try:
            with open(self.logfile, "a") as fout:
                subprocess.run(" ".join(lst), cwd=self.outdir, check=True, shell=True,
                            stdout = fout, stderr = fout)
        except subprocess.CalledProcessError:
            logging.info("GeneMark-ES in prediction mode has failed")
            touch(failpath)
        # predict and then clean
        self.gtf2faa()
        self.clean_gmes_files()

    def gtf2faa(self):
        lst = ["get_sequence_from_GTF.pl", "genemark.gtf", self.fasta]
        if not os.path.exists(self.gtf):
            logging.debug("There is no GTF file")
            return
        if os.path.exists(self.protfaa):
            logging.debug("Protein file already exists, skipping")
        else:
            try:
                with open(self.loggtf, "a") as fout:
                    subprocess.run(" ".join(lst), cwd=self.outdir, check=True, shell=True,
                                stdout = fout, stderr = fout)
            except subprocess.CalledProcessError:
                logging.warning("could not get proteins from gtf")
        # rename the proteins, to be compatibale with CAT
        #self.gtf2bed(self.gtf, self.bedfile)
        self.rename_for_CAT()

    def parse_gtf(self, gtf):
        """Given a gtf file from genemark es it extracts 
        some information to create a bed file"""
        nre = re.compile(r'gene_id "([0-9]+_g)\";')
        def beddict():
            return({"chrom": None, "r": [], "strand": None})
        beds = defaultdict(beddict)
        with open(gtf) as f:
            for line in f:
                # skip comment lines
                if line.startswith("#"):
                    continue

                l = line.split("\t")
                chrom = l[0].strip()
                start = int(l[3])
                stop = int(l[4])
                strand = l[6]

                # regex match
                m = (nre.findall(l[8]))
                if m is not None:
                    name = m[0]
                else:
                    continue
    
                # save all in the dictonary
                beds[name]["chrom"] = l[0]
                beds[name]["r"].append(int(l[3]))
                beds[name]["r"].append(int(l[4]))
                beds[name]["strand"] = l[6]
        return beds

    def gtf2bed(self, gtf, outfile, rename = None, beds = None):
        """
        given a faa file and a gtf(genemark-es format)
        we will be able to create a bed file, which can be used 
        with eukcc
        """
        if os.path.exists(outfile):
            logging.warning("Bedfile already exists, skipping")
            return


        # load gtf
        if beds is None:
            beds = self.parse_gtf(gtf)
        # write to file
        with open(outfile, "w") as f:
            for name, v in beds.items():
                if rename is not None:
                    name = rename[name]
                vals = "\t".join([v["chrom"], str(min(v["r"])), str(max(v["r"])), v['strand'], name])
                f.write("{}\n".format(vals))
        
    def rename_for_CAT(self, faa = None, gtf = None):
        """
        renames the protein file
        to matche the format:
            >contigname_ORFNUMBER

            eg:
                >NODE_1_1
        """
        self.finalfaa = os.path.join(self.outdir, "prot_final.faa")
        self.bedfile = os.path.join(self.outdir, "proteins.bed")
        if os.path.exists(self.finalfaa) and os.path.exists(self.bedfile):
            logging.debug("Renamed faa exists, skipping")
            return
        if faa is None:
            faa = self.protfaa
        if gtf is None:
            gtf = self.gtf
        try:
            faa = Fasta(faa)
        except FastaIndexingError:
            return
        except Exception as e:
            logging.debug("Unhandled pyfaidx Fasta error")
            print(e)
            return
        # load gtf
        beds = self.parse_gtf(gtf)
        orfcounter = defaultdict(int)
        # keep track of the renaming, so we can rename the bed
        renamed = {}
        # parse and rename
        with open(self.finalfaa, "w") as fout:
            for record in faa:
                if record.name not in beds.keys():
                    logging.warning("The protein was not found in the gtf file:")
                    print("protein: %s" % record.name)
                    print("GTF file: %s" % gtf)
                    logging.warning("stopping here, this is a bug in pygmes or an issue with GeneMark-ES")
                    exit()
                contig = beds[record.name]['chrom']
                orfcounter[contig] += 1
                # we use 1 as the first number, instead of the cool 0
                newprotname = "{}_{}".format(contig, orfcounter[contig])
                # keep track of the renaming, so we can rename the bed
                renamed[record.name] = newprotname
                fout.write(">{}\n{}\n".format(newprotname, record))
        # write renamed bed
        self.gtf2bed(self.gtf, self.bedfile, renamed, beds)

    def check_success(self):
        if self.finalfaa is False:
            return False
        if not os.path.exists(self.gtf):
            return False
        if not os.path.exists(self.finalfaa):
            return False

        # now more in detail
        # check if proteins are empty maybe
        with open(self.finalfaa) as fa:
            j = 1
            for line in fa:
                if j == 0:
                    if line.strip() == "":
                        return False
                    else:
                        return True
                j = j - 1

    def run_complete(self, models, diamonddb):
        self.selftraining()
        if self.check_success():
            logging.info("Ran GeneMark-ES successfully")
            # predict the lng now
            logging.info("Predicting the lng now")
            self.estimate_tax(diamonddb)
        else:
            logging.info("Using pre-trained models")
            self.fetchinfomap()
            self.premodel(models)
            if self.bestpremodel:
                self.bestpremodel.estimate_tax(diamonddb)
                self.premodeltax = self.bestpremodel.tax
                # print lineage of model compared to the infered tax
                print_lngs(self.modelinfomap[self.bestpremodel.modelname],
                           self.premodeltax)
                localmodals = self.infer_model(self.premodeltax)
                self.premodel(localmodals, stage=2)
                self.bestpremodel.estimate_tax(diamonddb)
                self.premodeltax = self.bestpremodel.tax
                # print lineage of model compared to the infered tax
                print_lngs(self.modelinfomap[self.bestpremodel.modelname],
                           self.premodeltax)
                # set the final values of of the protein prediction step
                self.finalfaa = self.bestpremodel.finalfaa
                self.bedfile = self.bestpremodel.bedfile
                self.tax = self.bestpremodel.tax
            # self.prediction()

    def estimate_tax(self, db):
        ddir = os.path.join(self.outdir, "diamond")
        create_dir(ddir)
        d = diamond(self.protfaa, ddir, db, sample=200, ncores = self.ncores)
        self.tax = d.lineage

    def premodel(self, models, stage=1):
        logging.debug("On bin: %s" % self.fasta)
        logging.debug("Running the pre Model stage %d" % stage)
        logging.debug("Using model directory: %s", models)
        self.bestpremodel = False
        modelfiles = glob.glob(os.path.join(models, "*.mod"))
        subgmes = []
        for model in modelfiles:
            logging.debug("Using model %s" % os.path.basename(model))
            name = os.path.basename(model)
            odir = os.path.join(self.outdir, "{}_premodels".format(stage), name)
            g = gmes(self.fasta, odir, ncores = self.ncores)
            g.prediction(model)
            if g.check_success():
                subgmes.append(g)

        if len(subgmes) == 0:
            logging.warning("Could not predict any proteins in this file")
        else:
            aminoacidcount = []
            for g in subgmes:
                fa = Fasta(g.protfaa)
                i = 0
                for seq in fa:
                    i += len(seq)
                aminoacidcount.append(i)
            # set the best model as the model leading to the most amino acids
            idx = aminoacidcount.index(max(aminoacidcount))
            self.bestpremodel = subgmes[idx]
            logging.info("Best model set as: %s" % os.path.basename(self.bestpremodel.model))

    def fetchinfomap(self):
        """
        function to make sure the information of all models
        is known to the class
        """
        if len(self.modelinfomap) == 0:
            info = self.fetch_info("{}info.csv".format(url))
            for line in info.split("\n"):
                l = line.split(",")
                if len(l) == 3:
                    self.modelinfomap[l[0]] = l[2].split("-")

    def infer_model(self, tax, n=3):
        """
        given we infered a lineage or we know a lineage
        we can try to fetch a model from the number of
        precomputed models that already exists
        for this we choose the model that shares the most
        taxonomic element with the predicted lineage
        If multiple modles have similar fit, we just again chose the best one
        """
        self.fetchinfomap()

        candidates = self.score_models(self.modelinfomap, tax)

        if len(candidates) > n:
            candidates = sample(candidates, n)
        # for each candidate, try to download the model into a file
        modeldir = os.path.join(self.outdir, "models")
        create_dir(modeldir)
        for model in candidates:
            self.fetch_model(modeldir, url, model)

        return modeldir

    def fetch_model(self, folder, url, name):
        url = "{}/models/{}.mod".format(url, name)
        modelfile = os.path.join(folder, "{}.mod".format(name))
        response = urllib.request.urlopen(url)
        data = response.read()  # a `bytes` object
        content = data.decode(
            "utf-8"
        )  # a `str`; this step can't be used if data is binary
        with open(modelfile, "w") as mod:
            mod.writelines(content)

    def fetch_info(self, url):
        response = urllib.request.urlopen(url)
        data = response.read()  # a `bytes` object
        infocsv = data.decode(
            "utf-8"
        )  # a `str`; this step can't be used if data is binary
        return infocsv

    def score_models(self, infomap, lng):
        scores = defaultdict(int)

        for model, mlng in infomap.items():
            for a, b in zip(lng, mlng):
                if int(a) == int(b):
                    scores[model] += 1
        # get all models with the highest score
        maxscore = max([v for k, v in scores.items()])
        candidates = []
        for m, s in scores.items():
            if s == maxscore:
                candidates.append(m)
        return candidates

    def writetax(self):
        """
        write infered taxonomy in a machine and human readble format
        """
        logging.info("Translating lineage")
        taxf = os.path.join(self.outdir, "lineage.txt")
        with open(taxf, "w") as fout:
           # get the information
            lng = self.tax
            nms = ncbi.get_taxid_translator(lng)
            ranks = ncbi.get_rank(lng)
            # first line is taxids in machine readable
            s = "-".join([str(i) for i in lng])
            fout.write("#taxidlineage: {}\n".format(s))
            fout.write("taxid\tncbi_rank\tncbi_name\n")
            for taxid in lng:
                if taxid in nms.keys():
                    name = nms[taxid]
                else:
                    name = "unnamed"
                fout.write(f"{taxid}\t{ranks[taxid]}\t{name}\n") 
        logging.info("Wrote lineage to %s" % taxf)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("pygmes/version.py") as fp:
    exec(fp.read(), version)
setuptools.setup(
    name="pygmes",
    version=version["__version__"],
    author="Paul Saary",
    author_email="saary@ebi.ac.uk",
    description="Run GeneMark-ES using pretrained models",
    url="https://github.com/openpaul/pygmes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    py_modules=["api"],
    entry_points={"console_scripts": ["pygmes = pygmes.api:main"]},
    install_requires=["ete3", "pyfaidx>=0.5.8"],
    packages=setuptools.find_packages(),
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
    ],
)
